<?php

namespace App\Admin;

use Sonata\AdminBundle\Admin\AbstractAdmin;
use Sonata\AdminBundle\Datagrid\DatagridMapper;
use Sonata\AdminBundle\Datagrid\ProxyQueryInterface;
use Sonata\AdminBundle\Datagrid\ListMapper;
use Sonata\AdminBundle\Form\FormMapper;
use Sonata\AdminBundle\Show\ShowMapper;
use Sonata\AdminBundle\Form\Type\ModelListType;
use Sonata\AdminBundle\Form\Type\ModelType;
use Symfony\Component\Security\Core\Authorization\AuthorizationCheckerInterface;
use Symfony\Component\Security\Core\Security;
use Doctrine\ORM\EntityManagerInterface;

final class CatalogosAdmin extends AbstractAdmin
{

    private $authorizationChecker;
    private $security;
    private $em;

    public function __construct(AuthorizationCheckerInterface $authorizationChecker, Security $security, EntityManagerInterface $em)
    {
        $this->authorizationChecker = $authorizationChecker;
        $this->security = $security;
        $this->em = $em;
    }

    protected function configureQuery(ProxyQueryInterface $query): ProxyQueryInterface
    {

        if ($this->authorizationChecker->isGranted('ROLE_ADMIN') || $this->authorizationChecker->isGranted('ROLE_SUPER_ADMIN')) {
            $query->orderBy($query->getRootAlias() . '.id', 'DESC');
        } elseif ($this->authorizationChecker->isGranted('ROLE_FABRICANTE')) {
            $user = $this->security->getUser();
            $fabricanteId = $user->getFabricante()->getId();
            $query
                ->andWhere($query->expr()->eq($query->getRootAlias() . '.fabricante', ':fabricanteId'))
                ->orderBy($query->getRootAlias() . '.id', 'DESC')
                ->setParameter('fabricanteId', $fabricanteId);
        }

        return $query;
    }

    protected function configureDatagridFilters(DatagridMapper $filter): void
    {
        $filter
            ->add('nombre')
            ->add('activo')
            ;
    }

    protected function configureListFields(ListMapper $list): void
    {
        $list
            ->add('nombre')
            ->add('descripcion')
            ->add('createdAt', null, ['format' => 'd/m/Y'])
            ->add('pdf', null, [
                'template' => 'admin/catalogo_pdf_list_field.html.twig'
            ])
            ->add('activo', null, ['editable' => true])
            ->add('_action', 'actions', [
                'actions' => [
                    'show' => [],
                    'edit' => [],
                ],
                'label' => 'Acciones',
            ]);
    }

    protected function configureFormFields(FormMapper $form): void
    {
        $user = $this->security->getUser();
        $fabricanteQueryBuilder = null;

        if ($this->security->isGranted('ROLE_SUPER_ADMIN') || $this->security->isGranted('ROLE_ADMIN')) {
            $fabricanteQueryBuilder = $this->em->getRepository('App\Entity\Fabricantes')
                ->createQueryBuilder('f');
        } else {
            $fabricante = $user->getFabricante();
            $fabricanteQueryBuilder = $this->em->getRepository('App\Entity\Fabricantes')
                ->createQueryBuilder('f')
                ->where('f.id = :fabricanteId')
                ->setParameter('fabricanteId', $fabricante->getId());
        }

        $form
            ->add('nombre')
            ->add('descripcion')
            ->add('fabricante', ModelType::class, [
                'class' => 'App\Entity\Fabricantes',
                'property' => 'nombre',
                'query' => $fabricanteQueryBuilder,
                'required' => true,
                'btn_add' => false,
                'label' => 'Fabricante',
            ])
            ->add('precio')
            ->add('medidas')
            ->add('pdf', ModelListType::class, [
                    'btn_add' => 'Añadir catálogo',
                    'btn_list' => 'Seleccionar',
                    'btn_delete' => 'Eliminar',
                ], [
                    'placeholder' => 'Catálogo no seleccionado',
                    'link_parameters' =>['context' => 'catalogo']
                ]
            )
        ;

        if ($this->authorizationChecker->isGranted('ROLE_SUPER_ADMIN') || $this->security->isGranted('ROLE_ADMIN')) {
            $form->add('activo');
        }
    }

    protected function configureShowFields(ShowMapper $show): void
    {
        $show
            ->with('Datos')
                ->add('nombre')
                ->add('descripcion')
                ->add('precio')
                ->add('medidas')
                ->add('fabricante.nombre', null, ['label' => 'Fabricante'])
            ->end()
            ->with('Vista')
                ->add('pdf', null, [
                    'label' => ' ',
                    'template' => 'admin/show_pdf.html.twig'
                ])
            ->end()
        ;
    }
}
